-- Hazara Khan's implementation of the Suicide Kings loot distribution
--  system. See http://www.wowwiki.com/Suicide_Kings for a description
--  of the system, and <Seraphim> of Medivh's loot rules for the basis
--  for this implementation.
-- This is intended to be a module for HazLoot, a master looter aid, but
--  written in such a way that it's testable outside of WoW. In theory.

if not HazLootList then -- not in WoW, because Ace creates HazLootList
  HSK = {}
  HSK.standalone = true
end

HSK.MasterList   = {} -- The master list    {[index] = memberName}
HSK.InvertedList = {} -- Inverted master list {[memberName]=index}
HSK.Alts         = {} -- Alts of members' mains in the master list
                      --  {[memberName]={alt1,alt2,...altN,}}
HSK.Mains        = {} -- Mains of members' alts (inverse of Alts)
HSK.Present      = {} -- A sub-list of the master, of present raiders

--- Add a new member to the Suicide Kings list at an (optional) position
-- Attempts to append a new member to the bottom of the list, or at the
--  position provided. If position would leave a gap at the bottom of the
--  list, then the member is added at the bottom, leaving no gaps.
-- @param member   The main character name of a raider in the SK List
-- @param position The position in the list of the new member. All existing
--                 members at, or below, this position will be shuffled down
--                 one slot. Position must be > 0
--                 If position is omitted, the new member starts at the bottom
-- @return The new position of the added member, nil if addition failed, or
--         member is nil.
function HSK:AddMember(member, position)
  if not member then return nil end
  if HSK.InvertedList[member] then
    return HSK.InvertedList[member]
  end
  local listsize = table.getn(HSK.MasterList)
  if nil == position or listsize < position then -- append
    table.insert(HSK.MasterList, member)
    position = table.getn(HSK.MasterList)
    HSK.InvertedList[member] = position
  elseif 0 < position and position <= listsize then
    for i = position,listsize do
      HSK.InvertedList[HSK.MasterList[i]] = i+1
    end
    table.insert(HSK.MasterList, position, member)
    HSK.InvertedList[member] = position
  else
    return nil
  end
  self:Debug("master list size now: "..table.getn(HSK.MasterList))
  return HSK.InvertedList[member]
end

--- Replace the master list with an ordered list of raiders.
-- @param members An array of members {member1, member2, ..., memberN,}, by
--                Suicide Kings order.
-- @return true if the list was created correctly
function HSK:CreateList(members)
  if not "table"==type(members) then return false end
  -- Check format of table
  -- HSK.MasterList = members
  -- return true
  return false
end

--- Remove a member from the Suicide Kings list
-- Permanently removes a member from the SK List - for example when they leave
-- the guild, stop raiding, are an utter douchebag, whatever. Adjusts the list
-- so that the list has no gaps.
-- @param member The name of a member to remove, or their position in the list.
-- @return true if the member was successfully removed
function HSK:RemoveMember(member)
  if not member then return false end
  local position
  local membername
  if "string" == type(member) then
    position = HSK:GetMasterPositionOf(member)
  elseif "number" == type(member) then
    position = member
  end
  if position >= 0 and position < table.getn(HSK.MasterList) then
    membername = HSK:GetMemberAt(position) -- may be an alt
    HSK.InvertedList[membername] = nil
    local listsize = table.getn(HSK.MasterList)
    for i = position+1,listsize do
      HSK.MasterList[i-1] = HSK.MasterList[i]
      HSK.InvertedList[HSK.MasterList[i]] = i-1
    end
    table.setn(HSK.MasterList, listsize-1)
    self:Debug("master list size now: "..table.getn(HSK.MasterList))
    return true
  end
  return false
end

--- Remove an alt from the records, leaving the main unaffected
--  @param alt The name of the alt in question
--  @return true if the alt was removed, false if the alt was a
--          main, nil if the alt was not found
function HSK:RemoveAlt(alt)
end

--- Promote an alt to main, and move the corresponding main to the alt list.
function HSK:PromoteAlt(alt)
end

--- Set an existing member to "present", and thus eligible for suicides
-- @param member The name of a member to mark as present
-- @return The member's position in the master list, nil if not found
function HSK:MarkAsPresent(member)
  return nil
end

--- Return a member's current position in the master SK List
--  @param position Integer position in the list.
--  @return The name of the member in the position provided, or nil if position
--          is larger than the size of the list.
function HSK:GetMemberAt(position)
  return HSK.MasterList[position]
end

--- Return the current position in the master list of a raid member
--  @param member A main in the master list, or one of their alts.
--  @return The position in the master list of a raid member, or the position in
--          the master list of a main of one of that member's alts. Returns nil
--          if member is not in the master list, nor an alt.
function HSK:GetMasterPositionOf(member)
  return HSK.InvertedList[member]
end

--- Suicide an active member, and adjust the master SK List.
--  @param member A main that has been marked as present, or one of their alts.
--  @return their new position in the list, nil if they were not found.
function HSK:Suicide(member)
  return nil
end

--- How many people are in the master list?
--  @return The size of the master list, 0 if empty.
function HSK:GetMasterListSize()
  return table.getn(HSK.MasterList)
end

--- How many people are in the "present" list?
--  @return The size of the present list, 0 if empty.
function HSK:GetPresentListSize()
  return table.getn(HSK.Present)
end

--- Return a (cloned) array of the master list.
--  The list returned is a copy of the master list. It should not therefore be
--  manipulated with any expectation of changes being reflected to the actual
--  list, but is appropriate for external persistence, for example, or display.
--  @return A cloned copy of the master SK List, which may be empty.
function HSK:GetMasterList()
  return nil
end

--- Return a (cloned) array of the list of members marked "present" for a raid.
--  The list returned is a copy. It should not therefore be manipulated with any
--  expectation of changes being reflected to the actual list, but is
--  appropriate for display, etc.
--  @return A cloned copy of the list of "present" raiders, in SK order, which
--          may be empty.
function HSK:GetPresentList()
  return nil
end

function HSK:Print(message)
  if HazLoot then
    HazLoot:Print(message)
  else
    print("[HSK] "..message)
  end
end

function HSK:Debug(message)
  if HazLoot then
    HazLoot:Debug(message)
  else
    print("[HSK][DEBUG] "..message)
  end
end

--- run tests
function HSK:Test()
  HSK:Debug("Starting SK List tests")
  local tmp, tmpPos
  assert(HSK:GetMasterListSize() == 0, "The Master list should start empty, but has non-zero size")
  assert(HSK:GetPresentListSize() == 0, "The Present list should start empty, but has non-zero size")

  --- AddMember
  HSK:Debug("AddMember()")
  assert(not HSK:AddMember(), "AddMember should return nil with no arguments")

  HSK:Debug("AddMember(\"TestMain-1\",-0) AddMember(\"TestMain-1\",-10)")
  assert(not HSK:AddMember("TestMain-1",0), "AddMember should return nil with position <= 0")
  assert(not HSK:AddMember("TestMain-1",-10), "AddMember should return nil with position <= 0")
  HSK:Debug("AddMember(\"TestMain1\")")
  tmp = HSK:AddMember("TestMain1")
  assert(tmp == 1, "The first member should be in position 1, but is "..(tmp or "nil"))

  HSK:Debug("AddMember(\"TestMain1\") (again)")
  tmp = HSK:AddMember("TestMain1")
  assert(tmp == 1, "The first member should still be in position 1, but is "..(tmp or "nil"))
  assert(HSK:GetMasterListSize() == 1, "The Master list should still have only 1 member, but has "..(tmp or "nil"))

  HSK:Debug("AddMember(\"TestMain2\",3)")
  tmp = HSK:AddMember("TestMain2",3)
  assert(tmp == 2, "The second member should be in position 2, but is "..(tmp or "nil"))
  assert(HSK:GetMasterListSize() == 2, "The Master list should now have 2 members, but has "..(tmp or "nil"))
  assert(HSK:GetPresentListSize() == 0, "The Present list still be empty, but has non-zero size")
  assert(1 == HSK:GetMasterPositionOf("TestMain1"), "TestMain1 should still be in position 1")
  assert(2 == HSK:GetMasterPositionOf("TestMain2"), "TestMain2 should still be in position 2")

  HSK:Debug("AddMember(\"TestMain3\",2)")
  tmp = HSK:AddMember("TestMain3",2)
  assert(2 == tmp, "TestMain3 should now be at position 2, but is "..(tmp or "nil"))
  tmp = HSK:GetMasterPositionOf("TestMain1")
  assert(1 == tmp, "TestMain1 should still be in position 1, but is "..(tmp or "nil"))
  tmp = HSK:GetMasterPositionOf("TestMain2")
  assert(3 == tmp, "TestMain2 should now be in position 3, but is "..(tmp or "nil"))

  --- RemoveMember
  HSK:Debug("RemoveMember(\"TestMain3\")")
  assert(HSK:RemoveMember("TestMain3"), "Failed to remove TestMain3")
  tmp = HSK:GetMasterPositionOf("TestMain3")
  assert(nil == tmp, "TestMain3 should have nil master position, not "..(tmp or "nil"))
  tmp = HSK:GetMasterPositionOf("TestMain1")
  assert(1 == tmp, "TestMain1 should still be in position 1, but is "..(tmp or "nil"))
  tmp = HSK:GetMasterPositionOf("TestMain2")
  assert(2 == tmp, "TestMain2 should now be back in position 2, but is "..(tmp or "nil"))
  HSK:Debug("AddMember(\"TestMain3\",2)")
  tmp = HSK:AddMember("TestMain3",2)

  HSK:Debug("RemoveMember(2)")
  assert(HSK:RemoveMember(2), "Failed to remove 2nd member, TestMain3")
  tmp = HSK:GetMasterPositionOf("TestMain3")
  assert(nil == tmp, "TestMain3 should have nil master position, not "..(tmp or "nil"))
  tmp = HSK:GetMasterPositionOf("TestMain1")
  assert(1 == tmp, "TestMain1 should still be in position 1, but is "..(tmp or "nil"))
  tmp = HSK:GetMasterPositionOf("TestMain2")
  assert(2 == tmp, "TestMain2 should now be back in position 2, but is "..(tmp or "nil"))

  HSK:Debug("Tests completed successfully")
end

if HSK.standalone then
  HSK:Test()
end
